from unittest import TestCase
import structlog_extensions.utils as utils


class Test_utils(TestCase):
    def test__parse_request_section(self):
        requests = {'get': 'GET /apache_pb.gif HTTP/1.0',
                    'post': 'POST /apache_pb.gif HTTP/1.0',
                    'put': 'PUT /apache_pb.gif HTTP/1.0',
                    'delete': 'DELETE /apache_pb.gif HTTP/1.0'}

        for method, request in requests.items():
            result = utils._parse_request_section(request)
            print(result)
            self.assertEqual(result['method'], method)

        request = "NotAValidRequestString"
        result = utils._parse_request_section(request)
        self.assertNotIn('method', result)

    def test__parse_user_agent_section(self):
        user_agents = {
            'Chrome': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/64.0.3282.186 Safari/537.36',
            'IE': 'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322)',
            'Firefox': 'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:54.0) Gecko/20100101 Firefox/54.0',
            'Googlebot-Image': 'Googlebot-Image/1.0',
            'Other': 'hjkhsdfjkhjrkf'}

        for agent_name, agent_string in user_agents.items():
            result = utils._parse_user_agent_section(agent_string)
            print(result)
            self.assertEqual(result['user_agent.name'], agent_name)





from unittest import TestCase
import structlog_extensions
import logging
import sys
import structlog

logging.basicConfig(
    format="%(message)s",
    stream=sys.stdout,
    level=logging.INFO,
)


class TestCombinedLogParser(TestCase):
    def setUp(self):
        self.logparser = structlog_extensions.processors.CombinedLogParser("gunicorn.access")
        self.logger = logging.Logger("gunicorn.access")
        self.method_name = "warning"
        self.valid_event_dict = {
            'event': '127.0.0.1 - - [05/Feb/2012:17:11:55 +0000] "GET / HTTP/1.1" 200 140 "-" "Mozilla/5.0 (Windows '
                     'NT 6.1; WOW64) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.5 Safari/535.19"'}
        self.invalid_event_dict = {
            'event': 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Quisque bibendum dolor nec arcu '
                     'posuere, eu egestas sapien lacinia. Pellentesque bibendum nulla non tellus finibus aliquet.'}
    def test_has_correct_status_code(self):
        self.assertEqual(200,
                         self.logparser(self.logger, self.method_name,
                                                           self.valid_event_dict)['http.response.status_code'],
                         "Status code not extracted correctly")
    def test_has_correct_host(self):
        self.assertEqual("127.0.0.1",
            self.logparser(self.logger, self.method_name, self.valid_event_dict)['source.ip'],
            "Host not extracted correctly")

    def test_has_correct_timestamp(self):
        self.assertEqual("2012-02-05T17:11:55+00:00",
            self.logparser(self.logger, self.method_name, self.valid_event_dict)['@timestamp'],
            "Timestamp incorrect")

    def test_method_correct(self):
        self.assertEqual("get",
            self.logparser(self.logger, self.method_name, self.valid_event_dict)['http.request.method'],
            "method incorrect")

    def test_url_correct(self):
        self.assertEqual("/",
            self.logparser(self.logger, self.method_name, self.valid_event_dict)[
                'url.original'],
            "url incorrect")

    def http_version_correct(self):
        self.assertEqual("1.1",
            self.logparser(self.logger, self.method_name, self.valid_event_dict)[
                'http.version'],
            "http version incorrect")

    # TODO: Finish off testing correctness of every field, also add asserts for presence of field

    def test_not_a_combined_log(self):
       self.assertNotIn('source.ip',self.logparser(self.logger, self.method_name, self.invalid_event_dict))

    def test_chain_processor(self):
        structlog.configure(
            processors=[
                structlog_extensions.processors.CombinedLogParser("access"),
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ],
            logger_factory=structlog.stdlib.LoggerFactory(),
        )

        formatter = structlog.stdlib.ProcessorFormatter(
            processor=structlog.dev.ConsoleRenderer(colors=False),
        )

        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        root_logger.setLevel(logging.INFO)

        logger = structlog.get_logger("access")
        logger.warning(
            '127.0.0.1 - frank [10/Oct/2000:13:55:36 -0700] "GET /apache_pb.gif HTTP/1.0" 200 2326 "http://www.example.com/start.html" "Mozilla/4.08 [en] (Win98; I ;Nav)"')





"""
structlog_extensions.processors

This module contains processors for structlog.
"""
import structlog
from .utils import convert_combined_log_to_ecs, unflatten_dict
import logging


class NestedDictJSONRenderer(structlog.processors.JSONRenderer):
    """
    Structlog processor for creating JSON output including nested key output. It assumes that any key name
    using the specified separator is actually multiple nested keys, and will transform the log event dictionary to
    reflect this. The primary use case is for turning
    'flat' key-value pairs into the sort of structured json seen in (for example) elastic common schema.

    Notes:
        Must be the last processor on a chain (when using structlog as a renderer),
        or the processor of a ``structlog.stdlib.ProcessorFormatter`` object when using logging as a renderer.

    Example:
        When using this logging initialisation:

        .. code-block:: python

            # --- std logging initialisation code using structlog rendering

            pre_chain = [
                # Add the log level and a timestamp to the event_dict if the log entry
                # is not from structlog.
                structlog.stdlib.add_log_level,
                structlog.stdlib.add_logger_name,
                structlog_extensions.processors.CombinedLogParser("gunicorn.access")
            ]

            logging.dict_config( {
                        "version": 1,
                        "disable_existing_loggers": False,
                        "formatters": {
                            "json_formatter": {
                                "()": structlog.stdlib.ProcessorFormatter,
                                "processor": structlog_extensions.processors.NestedDictJSONRenderer('.'),
                                "foreign_pre_chain": pre_chain,
                            }
                        },
                        "handlers": {
                            "error_console": {
                                "class": "logging.StreamHandler",
                                "formatter": "json_formatter",
                            },
                            "console": {
                                "class": "logging.StreamHandler",
                                "formatter": "json_formatter",
                            }
                        },
                    })

        These entries (produced by ``structlog_extensions.processors.CombinedLogParser``):

        .. code-block:: python

            { 'http.request.method': 'get', 'http:.request.referrer': 'http://www.example.com', 'http.version': '1.0'}`

        will be transformed into the following nested json structure:

        .. code-block:: python

            { 'http': { 'version': '1.0',
                        'request': { 'method': 'get',
                                     'referrer': 'http://www.example.com'}
                        }
            }

    Notes:
        In cases where a root key has a value assigned and potential subkeys exist, the behaviour of this processor
        could be difficult to predict since it uses a deep merge that will perform overwriting in the order the keys
        are supplied by the event_dict. To prevent this happening, use the `clean_fields` attribute to specify the
        conflicting keys that should be removed from the event_dict prior to expansion.

    Attributes:
        separator (str, optional): Namespace separator. Default = '_'
        clean_keys (list, optional): List of keys to remove from log event prior to expansion. Intended for use when the original
                             log event has key names that might overlap with the root names of nested keys.
    """

    def __init__(self, *args, clean_keys=None, separator='_', **kwargs):
        self.clean_fields = clean_keys
        self.separator = separator
        super().__init__(*args, **kwargs)

    def __call__(self, logger, name, event_dict):
        if self.clean_fields:
            for field in self.clean_fields:
                event_dict.pop(field, None)
        try:
            nested_dict = unflatten_dict(event_dict, self.separator)
        finally:
            return super().__call__(logger,name,nested_dict)


class CombinedLogParser:
    """
    Parses Apache combined log style entries in the event field into separate fields using
    `Elastic Common Schema <https://www.elastic.co/guide/en/ecs/current/ecs-field-reference.html>`_ field names.

    Attributes:
        target_logger (str): Name of the logger object that is logging combined log output.

    Example:
        Creating and using a parser instance with structlog:

        .. code-block:: python

            import structlog_extensions
            import structlog
            import logging



            structlog.configure(
                processors=[
                    structlog_extensions.processors.CombinedLogParser('gunicorn.access'),
                    structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
                ],
                logger_factory=structlog.stdlib.LoggerFactory(),
            )

            formatter = structlog.stdlib.ProcessorFormatter(
                processor=structlog.dev.ConsoleRenderer(colors=False),
            )

            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            root_logger = logging.getLogger()
            root_logger.addHandler(handler)
            root_logger.setLevel(logging.INFO)

            logger = structlog.get_logger("access")
            logger.warning(
                '127.0.0.1 - frank [10/Oct/2000:13:55:36 -0700] "GET /apache_pb.gif HTTP/1.0" 200 2326 "http://www.example.com/start.html" "Mozilla/4.08 [en] (Win98; I ;Nav)"')
    """
    def __init__(self, target_logger):
        self.target_logger = target_logger

    def __call__(self, logger, method_name, event_dict):
        try:
            record = event_dict.get("_record")
            if record is None:
                logger_name = logger.name
            else:
                logger_name = record.name
            if logger_name == self.target_logger:
                if method_name in ['info', 'warn', 'warning', 'error', 'critical', 'debug']:
                    severity = getattr(logging, method_name.upper())
                else:
                    severity = 0
                original_event = event_dict['event']
                ecs_fields = convert_combined_log_to_ecs(log_line=original_event, dataset=logger_name,
                                                         severity=severity)
                event_dict.update(ecs_fields)
        finally:
            return event_dict






from unittest import TestCase
import structlog
import structlog_extensions

class TestNestedDictJSONRenderer(TestCase):
    def test_chain_processor(self):
        structlog.configure(
            processors=[
                structlog_extensions.processors.NestedDictJSONRenderer(clean_keys=['event'], separator='_'),
            ],
            context_class=dict,
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )

        logger = structlog.get_logger(__name__)
        logger.warning('Test', event_original='Test', event_action='test', service_name='structlog_extensions', service_version='1.0.0' )




from setuptools import setup

KEYWORDS = ["logging", "structured", "structure", "log", "ecs", "apache-combined-log"]

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='structlog-extensions-nralbers',
    version='1.0.3',
    packages=['structlog_extensions'],
    url='https://structlog-extensions-nralbers.readthedocs.io/en/latest/',
    project_urls={
        'Documentation': 'https://structlog-extensions-nralbers.readthedocs.io/en/latest/',
        'Source': 'https://github.com/nralbers/structlog-extensions-nralbers',
        'Tracker': 'https://github.com/nralbers/structlog-extensions-nralbers/issues',
    },
    license='MIT',
    author='Niels Albers',
    author_email='nralbers@gmail.com',
    description='Processors for Structlog library',
    install_requires=['structlog','user-agents','deepmerge', 'pytz'],
    keywords=KEYWORDS,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    python_requires='~=3.5',
    classifiers=[
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers"
    ],
)



import re
from user_agents import parse
from deepmerge import  always_merger
import pytz
from datetime import datetime, timezone


_ecs_field_mappings = {'host':'source.ip',
                       'user': 'user.name',
                       'event': 'event.original',
                       'method': 'http.request.method',
                       'referrer': 'http.request.referrer',
                       'size': 'http.response.body.bytes',
                       'status': 'http.response.status_code',
                       'time': '@timestamp',
                       'url': 'url.original',
                       'version': 'http.version'}


def convert_combined_log_to_ecs(log_line, dataset, severity=0):
    """
    Converts a combined log entry into a dict containing the log entry key/values
    with the key names using Elastic Common schema element names.

    Args:
        log_line (str): Combined log entry
        dataset (str): source of the log entry (for example 'apache.access')
        severity (int, optional): severity of the source log event

    Returns:
        dict: Dictionary of key/value pairs with the key names using ECS namespaced names.
    """
    result = _parse_log_into_fields(log_line)
    request_fields = _parse_request_section(result['request'])
    result.update(request_fields)
    user_agent_fields = _parse_user_agent_section(result['agent'])
    message = '"{0}" {1} {2}'.format(result['request'], result['status'], result['size'])
    ecs_fields = _convert_field_names_to_ecs(result)
    ecs_fields.update(user_agent_fields)
    ecs_fields['message'] = message
    ecs_fields['event.original'] = log_line
    ecs_fields['event.dataset'] = dataset
    ecs_fields['ecs.version'] = '1.0.0'
    ecs_fields['event.created'] = datetime.now(timezone.utc).isoformat()
    ecs_fields['@timestamp'] = combined_log_timestring_to_iso(ecs_fields['@timestamp'])
    ecs_fields['event.severity'] = severity
    return ecs_fields


def _parse_log_into_fields(log_line):
    parts = [
        r'(?P<host>\S+)',  # host %h
        r'\S+',  # indent %l (unused)
        r'(?P<user>\S+)',  # user %u
        r'\[(?P<time>.+)\]',  # time %t
        r'"(?P<request>.+)"',  # request "%r"
        r'(?P<status>[0-9]+)',  # status %>s
        r'(?P<size>\S+)',  # size %b (careful, can be '-')
        r'"(?P<referrer>.*)"',  # referrer "%{Referrer}i"
        r'"(?P<agent>.*)"',  # user agent "%{User-agent}i"
    ]
    pattern = re.compile(r'\s+'.join(parts) + r'\s*\Z')
    match = pattern.match(log_line)
    if match:
        fields = match.groupdict()
        if fields["user"] == "-":
            fields["user"] = None
        fields["status"] = int(fields["status"])
        if fields["size"] == "-":
            fields["size"] = 0
        else:
            fields["size"] = int(fields["size"])
        if fields["referrer"] == "-":
            fields["referrer"] = None
        return fields
    else:
        return dict()


def _parse_request_section(request_string):
    request_matcher = r'(?P<method>\S+)\s+(?P<url>\S+)\s+HTTP/(?P<version>\S+)'
    pattern = re.compile(request_matcher)
    match = pattern.match(request_string)
    if match:
        request_fields = match.groupdict()
        if 'method' in request_fields:
            request_fields['method'] = request_fields['method'].lower()
        return request_fields
    else:
        return dict()


def _parse_user_agent_section(agent_string):
    user_agent = parse(agent_string)
    result = dict()
    result['user_agent.original'] = agent_string
    result['user_agent.name'] = user_agent.browser.family
    result['user_agent.os.name'] = user_agent.os.family
    result['user_agent.os.version'] = user_agent.os.version_string
    result['user_agent.os.full'] = ' '.join([user_agent.os.family, user_agent.os.version_string])
    result['user_agent.device.name'] = user_agent.device.family
    result['user_agent.version'] = user_agent.browser.version_string
    return result


def _convert_field_names_to_ecs(parsed_fields):
    ecs_fields = {_ecs_field_mappings[key]:value for (key,value) in parsed_fields.items() if key in _ecs_field_mappings}
    ecs_fields['event.action'] = ecs_fields['http.request.method']
    return ecs_fields

def unflatten_dict(flat_dict, separator='.'):
    """
    Turns a dict with key names defining a namespace into a nested dictionary

    Args:
        flat_dict (dict): The dict cont
        separator (str, optional): The separator used to split name elements. Default '.'

    Returns:
        dict: A nested dictionary structure created from the original flat dict.
    """
    expanded_dict = dict()
    for key, value in flat_dict.items():
        field_hierarchy = key.split(separator)
        for field in reversed(field_hierarchy):
            item = { field: value}
            value = item
        always_merger.merge(expanded_dict,value)
    return expanded_dict


def _parse_datetime(timestamp):
    '''
    Parses datetime with timezone formatted as:
        `day/month/year:hour:minute:second zone`

    Example:
        `>>> parse_datetime('13/Nov/2015:11:45:42 +0000')`
        `datetime.datetime(2015, 11, 3, 11, 45, 4, tzinfo=<UTC>)`

    Due to problems parsing the timezone (`%z`) with `datetime.strptime`, the
    timezone will be obtained using the `pytz` library.
    '''
    dt = datetime.strptime(timestamp[0:-6], '%d/%b/%Y:%H:%M:%S')
    dt_tz = int(timestamp[-5:-2]) * 60 + int(timestamp[-2:])
    return dt.replace(tzinfo=pytz.FixedOffset(dt_tz))


def combined_log_timestring_to_iso(timestamp):
    """
    converts a combined log format date/time entry into an iso standard datetime string
    Args:
        timestamp (str): Datetime log entry to convert

    Returns:
        str: iso standard datetime string
    """
    dt = _parse_datetime(timestamp)
    return dt.isoformat()

